import re
import sys

flower_pattern = re.compile(r"""([a-zA-Z0-9\- #]*)\t[kK]([cgioruvyCGIORUVY]*)[kK]""")
mutagen_pattern = re.compile(r"""(?:[a-zA-Z0-9\- #]+) *\t *([0-9]*[a-zA-Z]+)([0-9]+) *\t *([abAB]) *\t *([0-9]+) *\t *([0-9]+)""")
gene_pat = re.compile(r"""[cCoOgGyYiIuUrR]+""")
naming_pattern = re.compile(r"""([a-zA-Z0-9\- #]+)\t([a-zA-Z0-9\- #,]+)""")
flowers = []
mutagens = []

class Flower:
    def __init__(self, m):
        self.name = m.group(1)
        self.shortname = ""
        self.gene = m.group(2).upper()
        self.len = len(self.gene)
    def __str__(self):
        result = self.name + ":" + " "*(24-len(self.name))
        for i in range(self.len):
            result += " " + self.gene[i]
        return result
    def __cmp__(self, other):
        if isinstance(other, str):
            return cmp(self.name, other)
        elif isinstance(other, Flower):
            return cmp(self.name, other.name)
        else:
            return -1
    def hits(self,mute):
        # Returns two lists of pairs of the form (index,probability)
        result = [[],[]]
        for i in range(2):
            for j in range(self.len):
                upper = float(j+1)/self.len
                lower = float(j)/self.len
                if mute.max[i] > lower and mute.min[i] < upper:
                    bellow = lower-mute.min[i]
                    if bellow < 0:
                        bellow = 0
                    above = mute.max[i]-upper
                    if above < 0:
                        above = 0
                    result[i].append([(j+1), (1 - (bellow + above)/(mute.max[i]-mute.min[i]))])
        return result

class Mutagen:
    def __init__(self, region, num):
        self.region = region
        self.num = int(num)
        self.name = region + str(self.num)
        self.moss = ""
        self.max = [1.0]*2
        self.min = [0.0]*2
    def hit(self, ab, i, j):
        x = float(i)
        y = float(j)
        if ab == "a" or ab == "A":
            i = 0
        else:
            i = 1
        if self.max[i] > x/y:
            self.max[i] = x/y
        if self.min[i] < (x-1)/y:
            self.min[i] = (x-1)/y
    def __str__(self):
        result = self.moss + " " + self.name + ":"
        for i in range(2):
            result +="\t" + '%.2f'%(self.min[i]*100) + "% - " + '%.2f'%(self.max[i]*100) + "%\t"
        return result
    def __cmp__(self, other):
        if isinstance(other, str):
            result = cmp(self.name.lower(), other.lower())
        elif isinstance(other, Mutagen):
            result = cmp(self.region, other.region)
            if result == 0:
                result = cmp(self.num, other.num)
        else:
            result = -1
        return result
    def hits(self, flower):
        return flower.hits(self)

class Thread:
    def __init__(self, flower, index, history):
        self.flower = flower
        self.index = index
        self.history = history[:]
    def __str__(self):
        result = ""
        for i in range(len(self.history)):
            trip = self.history[i]
            if trip[2] == 0:
                result += "A"
            else:
                result += "B"
            result += "\t"
            if trip[0].shortname != "":
                result += trip[0].shortname
            else:
                result += trip[0].name
            result += "\t"
            if trip[1].moss != "":
                result += trip[1].moss
            else:
                result += trip[1].name
            if i != len(self.history)-1:
                result += "\t|\t"
        return result
    def addHistory(self, flower, mutagen, ab):
        self.history.append((flower,mutagen,ab))

def printTable():
    print "\n"
    print "|||| || ||||||||||||||  '''Sea Lillies'''  || ||||||||||  '''Ochids'''  || ||||||||||||||||  '''Roses of Ra'''  || ||||||||||||||  '''Sand Blooms'''  ||"
    print "|||| || ||  '''VA'''  ||  '''FR'''  ||  '''SI''' || '''BL'''  ||  '''CL'''  ||  '''EN'''  ||  '''MO'''  || ||  '''ST'''  ||  '''AN'''  || '''SL'''  ||  '''HD'''  ||  '''EU'''  || ||  '''BH'''  ||  '''WG'''  ||  '''GL'''  ||  '''PA'''  ||  '''PG'''  ||  '''EC'''  ||  '''OS'''  ||  '''NB'''  || ||  '''SU'''  ||  '''LD'''  ||  '''LT'''  ||  '''SA''' || '''CO''' ||  '''PM'''  ||  '''CA''' ||"
    print "||||  '''Mutagen'''  || ||  '''12'''  ||  '''18'''  ||  '''18''' || '''26'''  ||  '''36'''  ||  '''42'''  ||  '''44'''  || ||  '''28'''  ||  '''30'''  ||  '''31'''  ||  '''34'''  ||  '''36'''  || ||  '''08'''  ||  '''12'''  ||  '''12'''  ||  '''16'''  ||  '''17'''  ||  '''18'''  ||  '''25'''  ||  '''45'''  || ||  '''13'''  ||  '''18'''  ||  '''19'''  ||  '''24''' || '''26'''  ||  '''29'''  ||  '''34'''  || || '''Moss''' || '''Min%''' || '''Max%'''||"
    print "$TABLE l c i l l l l l l l i l l l l l i l l l l l l l l i l l l l l l l i l r r"
    lines = []
    for m in mutagens:
        for i in range(2):
            string = "|| " + m.name + " || "
            if i == 0:
                string += "A || "
            else:
                string += "B || "
            for f in flowers:
                hits = f.hits(m)
                for pairs in hits[i]:
                    if pairs[1] > .9999:
                        string += str(pairs[0]) + " " + f.gene[pairs[0]-1] + " "
                string += "|| "
            string +=  m.moss + " || "
            string += '%.2f'%(m.min[i]*100) + " || "
            string += '%.2f'%(m.max[i]*100) + " ||"
            lines.append([m.max[i],string])
    lines.sort()
    for l in lines:
        print l[1]
    print "\n"
    
def printMutagenHits(m):
    print m
    for i in range(2):
        if i == 0:    
            print "\nA:"
        else:
            print "\nB:"
        for f in flowers:
            hits = f.hits(m)
            string = " "*25
            for k in range(f.len):
                found = False
                for pair in hits[i]:
                    if (k+1) == pair[0]:
                        found = True
                if found:
                    string += " ^"
                else:
                    string += "  "
            string += "  "*(35-f.len)
            for pair in hits[i]:
                string += "(" + str(pair[0]) + ")" '%.2f'%(pair[1]*100) + "%  "
            print f
            print string

def printFlowerHits(flower):
    print ""
    possible = []
    definite = []
    for i in range(flower.len):
        possible.append([])
        definite.append([])
    for m in mutagens:
        hits = m.hits(flower)
        for i in range(2):
            if i == 0:
                ab = "A"
            else:
                ab = "B"
            for pair in hits[i]:
                if pair[1] > .9999:
                    definite[pair[0]-1].append(m.name + " " + ab)
                else:
                    possible[pair[0]-1].append("(" + m.name + " " + ab + " " + '%.2f'%(pair[1]*100) + "%)")
    for i in range(flower.len):
        string = str(i+1) + "\t" + flower.gene[i] + "\t"
        if len(definite[i]) > 0:
            for j in range(len(definite[i])):
                string += definite[i][j]
                if j != len(definite[i]) - 1:
                    string += ", "
        else:
            for j in range(len(possible[i])):
                string += possible[i][j]
                if j != len(possible[i]) - 1:
                    string += ", "
        print string
    print ""

def makeGene(gene):
    length = len(gene)
    flower = None
    for f in flowers:
        if f.len == length:
            flower = f
    if flower == None:
        print "Error: No genes of length", len
        return
    print "Base flower:", flower.name
    for i in range(length):
        if flower.gene[i] != gene[i]:
            t = Thread(flower,i,[])
            result = findGene(gene[i],[t])
            if result != None:
                print str(i+1) + ":\t" + str(result)
            else:
                print "Error get \"" + gene[i] + "\" to",i+1 
            
def findGene(target, threads):
    newThreads = []
    for t in threads:
        if t.flower.gene[t.index] == target:
            return t
        else:
            for m in mutagens:
                results = t.flower.hits(m)
                for i in range(2):
                    for pair in results[i]:
                        if pair[0]-1 == t.index and pair[1] > .9999:
                            for f in flowers:
                                opp = f.hits(m)
                                for opair in opp[(i+1)%2]:
                                    if opair[1] > .9999:
                                        newThread = Thread(f, opair[0]-1, t.history)
                                        newThread.addHistory(f,m,i)
                                        newThreads.append(newThread)
    if newThreads != []:
        return findGene(target, newThreads)
    else:
        return None
def readFiles():
    if len(sys.argv) > 1:
        f = open(sys.argv[1], "r")
    else:
        f = open("flowers.txt")
    print "Reading flower file..."
    line = f.readline()
    while(line != ''):
        match = flower_pattern.match(line)
        line = f.readline()
        if match != None:
            flowers.append(Flower(match))
    f.close()

    f = open("mutagens.txt")
    print "Reading mutagen file..."
    line = f.readline()
    while(line != ''):
        match = mutagen_pattern.match(line)
        line = f.readline()
        if match != None:
            name = match.group(1) + match.group(2)
            if name not in mutagens:
                mute = Mutagen(match.group(1), match.group(2))
                mutagens.append(mute)
            else:
                for m in mutagens:
                    if m == name:
                        mute = m
            mute.hit(match.group(3), match.group(4), match.group(5))
    f.close()
    mutagens.sort()

    f = open("names.txt")
    print "Reading names file..."
    line = f.readline()
    while(line != ''):
        match = naming_pattern.match(line)
        line = f.readline()
        if match != None:
            a = match.group(1)
            b = match.group(2)
            if a in mutagens:
                for m in mutagens:
                    if m == a:
                        m.moss = b
            elif a in flowers:
                for flower in flowers:
                    if flower == a:
                        flower.shortname = b
        else:
            print "Error: Unable to print line: \"" + line +"\""
    f.close()
    #flowers.sort()


def main():
    readFiles()
    for m in mutagens:
        for i in range(2):
            if m.max[i] <= m.min[i]:
                print "\nError:", m, "\n"

    while(1):
        for i in range(len(flowers)):
            print str(i+1)+")", flowers[i].name

        print "\nEnter a flower's number to see all mutagens which might hit a particular gene."
        print "Enter the name of a mutagen to see where it hits each flower."
        print "Type \"Print\" get a wiki formated print out of where each mutagen hits each flower."
        print "Enter a genome to see steps to turn a known flower into that genome."
        input = raw_input("> ")
        
        if input.isdigit() and int(input) > 0 and int(input) <= len(flowers):
            printFlowerHits(flowers[int(input)-1])
        elif input in mutagens:
            for m in mutagens:
                if m == input:
                    break
            printMutagenHits(m)
        elif input == "quit" or input == "exit":
            break
        elif input == "print":
            printTable()
        elif gene_pat.match(input):
            makeGene(input.upper())
        elif input != "":
            found = False
            for m in mutagens:
                if m.moss == input:
                    found = True
                    break
            if found:
                printMutagenHits(m)
            else:
                for f in flowers:
                    if f.shortname == input or f.name == input:
                        found = True
                        break
                if found:
                    printFlowerHits(f)
main()
